//'Hello World' netfilter hooks example
//For any packet, we drop it, and log fact to /var/log/messages

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <net/ip.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/udp.h>
#include <linux/init.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/hrtimer.h>
#include <linux/ktime.h>

#include <linux/errno.h>


#include "XtModasProc.h"
#include "XtModasNet.h"
#include "EbModas.h"

// struct holding set of hook function options
// static struct nf_hook_ops NfHookOptions;
#ifdef GPL
static struct hrtimer HrTimer;
#endif
// static struct socket *KernelSocket;

int TotalHRTCounter = 0;
int TotalHRTMissCounter = 0;

EXPORT_SYMBOL(TotalHRTCounter);
EXPORT_SYMBOL(TotalHRTMissCounter);

static struct socket *UdpServerSocket = NULL;
static DECLARE_COMPLETION( ThreadComplete );
static struct task_struct *TaskUDPServer;
#define SERVER_PORT 17784

#define PACKET_DUMP_LEN       32
#define MS_TO_NS(x)           (x * 1E6L)
#define COUNT_HRT_TO_DEBUG    1000

#include <linux/netpoll.h>
#define MESSAGE_SIZE 1024
#define INADDR_LOCAL ((unsigned long int)0xc0a8BE04) //192.168.190.4
#define INADDR_SEND ((unsigned long int)0xc0a8BE3B) //192.168.190.59
static struct netpoll* np = NULL;
static struct netpoll np_t;

unsigned int XtModasPacketCallback(const struct nf_hook_ops *ops, struct sk_buff *skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
#if PACKET_DUMP_LEN > 0
  int i;
  if (skb_mac_header_was_set(skb))
  {
    char *Mac = skb_mac_header(skb);
    if (!(Mac[12] == 0x08 && Mac[13] == 0x00))
    {
      if (in)
      {
        printk(KERN_INFO "%s (In): ", in->name);
      }
      if (out)
      {
        printk(KERN_INFO "%s (Out): ", out->name);
      }
//      printk(KERN_INFO "(Packet %p -> %p) ", in, out);
      for (i = 0; i < PACKET_DUMP_LEN; i++)
      {
        printk("%02X%c", Mac[i], i == (PACKET_DUMP_LEN-1) ? '\n' : ':');
      }
    }
  }
#endif
  return NF_ACCEPT;
}

#if 0
struct task_struct *task;
int ThreadDataPointer;
int XtModasThread(void *data)
{
  struct kvec iov;
  int ThreadLoopCounter = 0;
  allow_signal(SIGKILL);
  do
  {
    unsigned int s;
    struct msghdr msg;

    iov.iov_base = kmalloc(2048, GFP_KERNEL);
    iov.iov_len = 2048;

    memset(&msg, 0, sizeof(msg));
    s = kernel_recvmsg(KernelSocket, &msg, &iov, 1, iov.iov_len, 0);
//    printk(KERN_INFO "Received %d Bytes\n", s);

    if (s < 0)
    {
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout(10000);
      ThreadLoopCounter++;
      if (ThreadLoopCounter >= 1000)
      {
        // printk(KERN_INFO "Count overflowed 1000 -> Reset\n");
        ThreadLoopCounter = 0;
      }
    }
    kfree(iov.iov_base);
  } while(!kthread_should_stop());
  printk(KERN_INFO "Kernel thread exit\n");
  return 0;
}
#endif

#ifdef GPL
enum hrtimer_restart HrTimerCB(struct hrtimer *timer)
{
  static int HRTCounter = 0;
  static int HRTMissCounter = 0;

  int TimerMisses;
  ktime_t TimeToFire, TimerNow;
  unsigned long delay_in_ms = 1L;
  TimeToFire = ktime_set(0, MS_TO_NS(delay_in_ms));
  TimerNow = ktime_get();
  TimerMisses = hrtimer_forward(timer, TimerNow, TimeToFire);

  if (TimerMisses > 1)
  {
    HRTMissCounter += TimerMisses - 1;
  }
  HRTCounter++;
  if ((HRTCounter + HRTMissCounter) >= COUNT_HRT_TO_DEBUG)
  {
#if 0
    printk(KERN_INFO "HRT %d*%dms %d missed\n", HRTCounter, (int) delay_in_ms, HRTMissCounter);
#endif
    TotalHRTMissCounter += HRTMissCounter;
    TotalHRTCounter += HRTCounter;
    HRTCounter = 0;
    HRTMissCounter = 0;
  }
  return HRTIMER_RESTART;
}
#endif

// #define ETH_E_PROFINET  0x8892
#define ETH_E_PROFINET  0x0800

unsigned char RxPacketBuffer[2048];
char TxPacketBuffer[2048];
static int UdpServerThread(void *data)
{
  struct sockaddr_in client;
  struct sockaddr *SenderAddress;
  int len;
  struct msghdr msg;
  struct iovec iov;
  mm_segment_t oldfs;

  allow_signal(SIGTERM);
  if (UdpServerSocket->sk == NULL)
    return 0;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  msg.msg_iov  = &iov;
  msg.msg_iovlen = 1;
  msg.msg_namelen = sizeof(struct sockaddr_in);
  msg.msg_iov  = &iov;
  msg.msg_name = &client;
  msg.msg_flags = 0;
  while (!signal_pending(current))
  {
    iov.iov_base = RxPacketBuffer;
    iov.iov_len  = 50;
    oldfs = get_fs();
    set_fs(KERNEL_DS);
    len = sock_recvmsg(UdpServerSocket, &msg, sizeof(RxPacketBuffer), 0);
    set_fs(oldfs);
    if (len > 0)
    {

      unsigned short __attribute__((__unused__)) FromPort = 0;
      SenderAddress = (struct sockaddr *)&client;
      FromPort = (SenderAddress->sa_data[0]) | (SenderAddress->sa_data[1] << 8);
#if 0
      printk(KERN_INFO "Rx UDP From %u.%u.%u.%u:%d\n",
             (unsigned char) SenderAddress->sa_data[2],
             (unsigned char) SenderAddress->sa_data[3],
             (unsigned char) SenderAddress->sa_data[4],
             (unsigned char) SenderAddress->sa_data[5], htons(FromPort));
#endif
#if 0
      {
        int len;
        struct sockaddr_in to;
        struct msghdr txmsg;
        struct iovec txiov;

        to.sin_family = AF_INET;
        memcpy(&to.sin_addr.s_addr, &SenderAddress->sa_data[2], 4);
        to.sin_port = FromPort + 1;

        memcpy(TxPacketBuffer, "hallo", 6);
        txmsg.msg_control = NULL;
        txmsg.msg_controllen = 0;
        txmsg.msg_iov  = &txiov;
        txmsg.msg_iovlen = 1;
        txmsg.msg_name = &to;
        txmsg.msg_namelen = sizeof(to);
        txmsg.msg_flags = MSG_DONTWAIT;
        txiov.iov_base = TxPacketBuffer;
        txiov.iov_len = 512;

        oldfs = get_fs();
        set_fs(KERNEL_DS);
        len = sock_sendmsg(UdpServerSocket, &txmsg, 512);
        if (len)
        {
          int i;
          for (i = 0; i < 20; i++)
            sock_sendmsg(UdpServerSocket, &txmsg, 512);
        }
      }
      if (len < 0)
        printk(KERN_ERR "sock_sendmsg returned: %d\n", len);
      set_fs(oldfs);
#endif
      if (np)
      {
        int i;
        for (i = 0; i < 20; i++)
          netpoll_send_udp(np, TxPacketBuffer, 512);
      }
    }
  }
  complete(&ThreadComplete);
  return 0;
}


// Siehe auch: http://cs.uccs.edu/~smanet/aodvsrc/utils.c
// http://www.linuxquestions.org/questions/programming-9/socket-programming-in-kernel-tcp-ip-403685/
// http://www.hep.by/gnu/kernel/networking/API-kernel-recvmsg.html
static int InitXtModas(void)
{
  struct sockaddr_in server;
  int servererror;

  XtModasNetInit();

  if (sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &UdpServerSocket) < 0)
  {
    printk( KERN_ERR "server: Error creating udpsocket.\n" );
    return -EIO;
  }
  UdpServerSocket->sk->sk_reuse = 1;
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons( (unsigned short)SERVER_PORT );
  servererror = UdpServerSocket->ops->bind( UdpServerSocket,
      (struct sockaddr *) &server, sizeof( server ) );
  if (servererror)
  {
    sock_release( UdpServerSocket );
    return -EIO;
  }

  TaskUDPServer = kthread_run(UdpServerThread, NULL, "UDPServer");
  if (!TaskUDPServer)
  {
    sock_release(UdpServerSocket);
    return -EIO;
  }

#if 0
  np_t.name = "LRNG";
  strlcpy(np_t.dev_name, "eth0", IFNAMSIZ);
  np_t.local_ip.ip = htonl(INADDR_LOCAL);
  np_t.remote_ip.ip = htonl(INADDR_SEND);
  np_t.local_port = 6665;
  np_t.remote_port = 6666;
  memset(np_t.remote_mac, 0xff, ETH_ALEN);
  netpoll_print_options(&np_t);
  netpoll_setup(&np_t);
  np = &np_t;
#endif

#if 0
  int RC;
#ifdef GPL
  ktime_t TimeToFire;
#endif

  RC = sock_create_kern(AF_PACKET, SOCK_RAW, __constant_htons(ETH_E_PROFINET), &KernelSocket);
  if (RC)
  {
    printk(KERN_INFO "Create socket failed: %d\n", RC);
  } else if (KernelSocket)
  {
    struct net_device *RawNetDev;
    struct sockaddr_ll sll;
    printk(KERN_INFO "Created raw socket 0x%p\n", KernelSocket);
#if 1
    lock_sock(KernelSocket->sk);
    KernelSocket->sk->sk_sndbuf = PAGE_SIZE * 2;
    KernelSocket->sk->sk_rcvbuf = PAGE_SIZE * 2;
    KernelSocket->sk->sk_write_space(KernelSocket->sk);
    release_sock(KernelSocket->sk);
#else
    int r;
    int SocketBufferSize = PAGE_SIZE;
    r = sock_setsockopt(KernelSocket, SOL_SOCKET, SO_SNDBUF, (void *) &SocketBufferSize, sizeof (int));
    if (r < 0)
    {
      printk(KERN_ERR "Error setting buffsize %d (error: %d)\n", SocketBufferSize, r);
    } else
    {
      printk(KERN_INFO "Socket buffer size set to %d\n", SocketBufferSize);
    }
#endif
    RawNetDev = dev_get_by_name(&init_net, "eth0");

    memset(&sll, 0, sizeof(sll));
    sll.sll_family = AF_PACKET;
    sll.sll_ifindex = RawNetDev->ifindex;
    sll.sll_protocol = __constant_htons(ETH_E_PROFINET);

    if (KernelSocket->ops->bind(KernelSocket, (struct sockaddr *) &sll, sizeof(sll)) < 0)
    {
      printk(KERN_ERR "bind(ETH_E_PROFINET) failed");
    }
  }

#if 0
  ThreadDataPointer = 20;
  task = kthread_run(&XtModasThread, (void *)ThreadDataPointer, "XtModas");
  printk(KERN_INFO "Started kernel thread: %s\n", task->comm);
#endif


  NfHookOptions.hook = XtModasPacketCallback;           // function to call when conditions below met
  NfHookOptions.hooknum = NF_INET_PRE_ROUTING;          // called right after packet recieved, first hook in Netfilter
  NfHookOptions.pf = NFPROTO_IPV4 /* NFPROTO_ARP */;    // IPV4 packets
  NfHookOptions.priority = NF_IP_PRI_FIRST;             // set to highest priority over all other hook functions
  if (nf_register_hook(&NfHookOptions))                 // register hook
  {
    printk(KERN_ERR "Register netfilter hook failed\n");
  } else
  {
    printk(KERN_INFO "Registered netfilter hook\n");
  }
  XtModasProcInit();

#ifdef GPL
  printk("Startinged HR-Timer\n");
  TimeToFire = ktime_set(0, MS_TO_NS(1L));
  hrtimer_init( &HrTimer, CLOCK_MONOTONIC, HRTIMER_MODE_REL );
  HrTimer.function = &HrTimerCB;
  hrtimer_start( &HrTimer, TimeToFire, HRTIMER_MODE_REL );
#endif
#endif
  // EbModasInit();

  return 0;                                             // return 0 for success
}

void ExitXtModas(void)
{
#ifdef GPL
  int RC;
#endif
  // EbModasExit();

  if (TaskUDPServer)
  {
    send_sig(SIGTERM, TaskUDPServer, 0);
    wait_for_completion(&ThreadComplete);
  }
  if (UdpServerSocket)
    sock_release(UdpServerSocket);


#if 0

  nf_unregister_hook(&NfHookOptions);                   // cleanup – unregister hook
  kthread_stop(task);
  kernel_sock_shutdown(KernelSocket, SHUT_WR);
  sock_release(KernelSocket);
  KernelSocket = NULL;
#ifdef GPL
  RC = hrtimer_cancel( &HrTimer );
  if (RC) printk("The timer was still in use...\n");
  printk("HR Timer module uninstalling\n");
#endif
  XtModasProcExit();
#endif
  XtModasNetDeInit();
}

module_init(InitXtModas);
module_exit(ExitXtModas);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Graf <tg@modas.de>");
MODULE_DESCRIPTION("-");
MODULE_ALIAS("ipt_modas");

