/*
 * XtModasNet.c
 *
 *  Created on: 06.01.2014
 *      Author: thomas
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/etherdevice.h>
#include <linux/pci.h>

#define MODAS_NETWORK_IF

struct NetDevPrivateData_s
{
  struct net_device_stats NetDevStats;
};

static int NetDevGetOpen(struct net_device *dev)
{
  printk(KERN_INFO "%s: NetDevGetOpen\n", dev->name);
  return 0;
}

static int NetDevGetStop(struct net_device *dev)
{
  printk(KERN_INFO "%s: NetDevGetStop\n", dev->name);
  return 0;
}

static int NetDevGetStartXmit(struct sk_buff *skb, struct net_device *dev)
{
  struct NetDevPrivateData_s *NDPD = netdev_priv(dev);
  NDPD->NetDevStats.tx_packets++;
  NDPD->NetDevStats.tx_bytes += skb->len;
  return 0;
}

static struct net_device_stats* NetDevGetStats(struct net_device *dev)
{
  struct NetDevPrivateData_s *NDPD = netdev_priv(dev);
//  printk(KERN_INFO "NetDevGetStats is called\n");
  return &NDPD->NetDevStats;
}

static const struct net_device_ops NetDevOps =
{
 .ndo_open = NetDevGetOpen,
 .ndo_stop = NetDevGetStop,
 .ndo_start_xmit = NetDevGetStartXmit,
 .ndo_get_stats = NetDevGetStats,
};

#ifdef MODAS_NETWORK_IF
static struct net_device *NetDev;
static struct pci_dev *PciDev;
#endif

int XtModasNetInit(void)
{
#ifdef MODAS_NETWORK_IF
//  static struct net_device *TestDev = __dev_get_by_name(net, ifr->ifr_name);

  PciDev = pci_get_device(0x16C3, 0xABCD, NULL);
  if (PciDev)
  {
    printk(KERN_INFO "pci: Device found\n");
  } else
  {
    printk(KERN_INFO "pci: Device not found\n");
  }

  /*
   * alloc_etherdev allocates memory for dev and dev->priv.
   * dev->priv shall have sizeof(struct rtl8139_private) memory
   * allocated.
   */
  NetDev = alloc_etherdev(sizeof(struct NetDevPrivateData_s));
  if (NetDev)
  {
    int i;
    /* UPDATE NET_DEVICE */
    for(i = 0; i < 6; i++)
    {
      /* Hardware Address */
      NetDev->dev_addr[i] = i;
      NetDev->broadcast[i] = 0xff;
    }
    NetDev->hard_header_len = 14;
    NetDev->netdev_ops = &NetDevOps;
    if (register_netdev(NetDev))
    {
      printk(KERN_ERR "Could not register netdevice\n");
    }
  }

#endif
  return 0;
}

int XtModasNetDeInit(void)
{
#ifdef MODAS_NETWORK_IF
  unregister_netdev(NetDev);
#endif
  return 0;
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Graf <tg@modas.de>");
