#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include "XtModasProc.h"

#define MESSAGE_LENGTH 256
static char TmpMessageBuffer[MESSAGE_LENGTH];

static struct proc_dir_entry *ProcDirEntry;
#define PROC_ENTRY_FILENAME "modas"

#define MIN(a,b)  ((a > b) ? b : a)

static ssize_t ProcRead(struct file *File,          /* see include/linux/fs.h   */
                        char *DstBuffer,            /* The buffer to put data to (in the user segment)    */
                        size_t DstBufferLen,        /* The length of the buffer */
                        loff_t *Offset)
{
  static int ReadFinished = 0;
  int i;
  char *OutputMessage = NULL;

  if (ReadFinished)
  {
    ReadFinished = 0;
    return 0;
  }

  OutputMessage = kmalloc(MESSAGE_LENGTH + 30, GFP_KERNEL);
  sprintf(OutputMessage,
          "Last input: %s\nHRT counter: %d\nMissed HRT Events %d\n",
          TmpMessageBuffer,
          TotalHRTCounter,
          TotalHRTMissCounter);
  {
    int CopyLen = MIN(DstBufferLen, strlen(OutputMessage));
    for (i = 0; i < CopyLen; i++)
    {
      put_user(OutputMessage[i], DstBuffer + i);
    }
  }
  ReadFinished = 1;
  return i;   /* Return the number of bytes "read" */
}

/*
 * This function receives input from the user when the user writes to the /proc
 * file.
 */
static ssize_t ProcWrite(struct file *File,  /* The file itself */
                         const char *Buffer,  /* The buffer with input */
                         size_t Length,  /* The buffer's length */
                         loff_t *Offset)
{       /* offset to file - ignore */
  int i;

  /*
   * Put the input into Message, where module_output will later be
   * able to use it
   */
  for (i = 0; i < MESSAGE_LENGTH - 1 && i < Length; i++)
  {
    get_user(TmpMessageBuffer[i], Buffer + i);
  }
  /*
   * we want a standard, zero terminated string
   */
  TmpMessageBuffer[i] = '\0';

  /*
   * We need to return the number of input characters used
   */
  return i;
}

/*
 * 1 if the file is currently open by somebody
 */
int AlreadyOpen = 0;

/*
 * Queue of processes who want our file
 */
DECLARE_WAIT_QUEUE_HEAD(WaitQ);

/*
 * Called when the /proc file is opened
 */
static int ProcOpen(struct inode *ProcInode, struct file *ProcFile)
{
  /*
   * If the file's flags include O_NONBLOCK, it means the process doesn't
   * want to wait for the file.  In this case, if the file is already
   * open, we should fail with -EAGAIN, meaning "you'll have to try
   * again", instead of blocking a process which would rather stay awake.
   */
  if ((ProcFile->f_flags & O_NONBLOCK) && AlreadyOpen)
  {
    return -EAGAIN;
  }

  /*
   * This is the correct place for try_module_get(THIS_MODULE) because
   * if a process is in the loop, which is within the kernel module,
   * the kernel module must not be removed.
   */
  try_module_get(THIS_MODULE);

  /*
   * If the file is already open, wait until it isn't
   */

  while (AlreadyOpen)
  {
    int i, is_sig = 0;

    /*
     * This function puts the current process, including any system
     * calls, such as us, to sleep.  Execution will be resumed right
     * after the function call, either because somebody called
     * wake_up(&WaitQ) (only module_close does that, when the file
     * is closed) or when a signal, such as Ctrl-C, is sent
     * to the process
     */
    wait_event_interruptible(WaitQ, !AlreadyOpen);

    /*
     * If we woke up because we got a signal we're not blocking,
     * return -EINTR (fail the system call).  This allows processes
     * to be killed or stopped.
     */

    /*
     * Emmanuel Papirakis:
     *
     * This is a little update to work with 2.2.*.  Signals now are contained in
     * two words (64 bits) and are stored in a structure that contains an array of
     * two unsigned longs.  We now have to make 2 checks in our if.
     *
     * Ori Pomerantz:
     *
     * Nobody promised me they'll never use more than 64 bits, or that this book
     * won't be used for a version of Linux with a word size of 16 bits.  This code
     * would work in any case.
     */
    for (i = 0; i < _NSIG_WORDS && !is_sig; i++)
    {
      is_sig = current->pending.signal.sig[i] & ~current->blocked.sig[i];
    }

    if (is_sig)
    {
      /*
       * It's important to put module_put(THIS_MODULE) here,
       * because for processes where the open is interrupted
       * there will never be a corresponding close. If we
       * don't decrement the usage count here, we will be
       * left with a positive usage count which we'll have no
       * way to bring down to zero, giving us an immortal
       * module, which can only be killed by rebooting
       * the machine.
       */
      module_put(THIS_MODULE);
      return -EINTR;
    }
  }

  /*
   * Open the file
   */
  AlreadyOpen = 1;
  return 0;   /* Allow the access */
}

/*
 * Called when the /proc file is closed
 */
static int ProcClose(struct inode *ProcInode, struct file *ProcFile)
{
  /*
   * Set Already_Open to zero, so one of the processes in the WaitQ will
   * be able to set Already_Open back to one and to open the file. All
   * the other processes will be called when Already_Open is back to one,
   * so they'll go back to sleep.
   */
  AlreadyOpen = 0;

  /*
   * Wake up all the processes in WaitQ, so if anybody is waiting for the
   * file, they can have it.
   */
  wake_up(&WaitQ);
  module_put(THIS_MODULE);
  return 0;   /* success */
}

/*
 * File operations for our proc file. This is where we place pointers to all
 * the functions called when somebody tries to do something to our file. NULL
 * means we don't want to deal with something.
 */
static struct file_operations ProcFileOps =
{
 .read = ProcRead,        /* "read" from the file */
 .write = ProcWrite,      /* "write" to the file */
 .open = ProcOpen,        /* called when the /proc file is opened */
 .release = ProcClose,    /* called when it's closed */
};

/*
 * Inode operations for our proc file.  We need it so we'll have somewhere to
 * specify the file operations structure we want to use, and the function we
 * use for permissions. It's also possible to specify functions to be called
 * for anything else which could be done to an inode (although we don't bother,
 * we just put NULL).
 */
int XtModasProcInit(void)
{
  printk(KERN_INFO "Modas: Init\n");
  ProcDirEntry = proc_create(PROC_ENTRY_FILENAME, 0644, NULL, &ProcFileOps);
  return 0;
}

void XtModasProcExit(void)
{
  proc_remove(ProcDirEntry);
  printk(KERN_INFO "Modas: Exit\n");
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Graf <tg@modas.de>");
MODULE_DESCRIPTION("-");
